import argparse

import rich
from ansible import context
from ansible.cli.arguments import option_helpers as opt_help
from ansible.errors import AnsibleOptionsError, AnsibleUndefinedVariable
from ansible.module_utils._text import to_native
from ansible.utils.display import Display

from ansible_variables import __version__
from ansible_variables.cli import CLI
from ansible_variables.utils.vars import variable_sources

display = Display()

# Internal vars same as defined for ansible-inventory
# pylint: disable=line-too-long
# (https://github.com/ansible/ansible/blob/d081ed36169f4f74512d1707909185281a30e29b/lib/ansible/cli/inventory.py#L28-L46
INTERNAL_VARS = frozenset(
    [
        "ansible_diff_mode",
        "ansible_config_file",
        "ansible_facts",
        "ansible_forks",
        "ansible_inventory_sources",
        "ansible_limit",
        "ansible_playbook_python",
        "ansible_run_tags",
        "ansible_skip_tags",
        "ansible_verbosity",
        "ansible_version",
        "inventory_dir",
        "inventory_file",
        "inventory_hostname",
        "inventory_hostname_short",
        "groups",
        "group_names",
        "omit",
        "playbook_dir",
    ]
)


class AnsibleVariablesVersion(argparse.Action):
    """we want to have our ansible-variables package version in the --version output"""

    def __call__(self, parser, namespace, values, option_string=None):
        ansible_version = to_native(opt_help.version(f"ansible-variables {__version__}"))
        print(ansible_version)
        parser.exit()


class VariablesCLI(CLI):
    """used to display from where a variable value is coming from"""

    name = "ansible-variables"

    def __init__(self, args):
        super().__init__(args)
        self.loader = None
        self.inventory = None
        self.vm = None  # pylint: disable=invalid-name

    def init_parser(self, usage="", desc=None, epilog=None):
        super().init_parser(
            usage="usage: %prog [options] [host]",
            epilog="""Show variable sources for a host.
                    Copyright 2023, Christoph Hille, https://github.com/hille721/ansible-variables.""",
        )
        version_help = (
            "show program's version number, config file location, configured module search path,"
            " module location, executable location and exit"
        )

        self.parser.add_argument("--version", action=AnsibleVariablesVersion, nargs=0, help=version_help)

        opt_help.add_inventory_options(self.parser)
        opt_help.add_vault_options(self.parser)
        opt_help.add_basedir_options(self.parser)

        # remove unused default options
        self.parser.add_argument("--list-hosts", help=argparse.SUPPRESS, action=opt_help.UnrecognizedArgument)
        self.parser.add_argument(
            "-l",
            "--limit",
            help=argparse.SUPPRESS,
            action=opt_help.UnrecognizedArgument,
        )

        self.parser.add_argument(
            "host",
            action="store",
            help="Ansible hostname for which variable sources should be printed",
        )

        self.parser.add_argument(
            "--var",
            action="store",
            default=None,
            dest="variable",
            help="Only check for specific variable",
        )

    def post_process_args(self, options):
        options = super().post_process_args(options)

        display.verbosity = options.verbosity
        self.validate_conflicts(options)

        return options

    def run(self):
        super().run()

        # Initialize needed objects
        self.loader, self.inventory, self.vm = self._play_prereqs()
        verbosity = display.verbosity

        host = self.inventory.get_host(context.CLIARGS["host"])
        if not host:
            raise AnsibleOptionsError("You must pass a single valid host to ansible-variables")

        for variable in variable_sources(
            variable_manager=self.vm,
            host=host,
            var=context.CLIARGS["variable"],
        ):
            if variable.name not in INTERNAL_VARS:
                if context.CLIARGS["variable"] and not variable.value:
                    # variable as options passed which is not defined
                    raise AnsibleUndefinedVariable("Variable %s is not defined" % variable.name)
                rich.print(
                    f"[bold]{variable.name}[/bold]: {variable.value} - [italic]{variable.source_mapped}[/italic]"
                )
                if verbosity >= 1:
                    files = variable.file_occurrences(loader=self.loader)
                    for ffile in files:
                        rich.print(ffile)


def main(args=None):
    VariablesCLI.cli_executor(args)


if __name__ == "__main__":
    main()



import pytest
from ansible import constants as C
from ansible.errors import AnsibleFileNotFound
from ansible.parsing.dataloader import DataLoader

from ansible_variables.utils.vars import VariableSource

DEBUGLOG = """
29355 1681471136.41205:  29355 1681471136.40830: in VariableManager get_vars()
29355 1681471136.40839: Calling all_inventory to load vars for server1
29355 1681471136.40842: Calling groups_inventory to load vars for server1
29355 1681471136.40850: Calling all_plugins_inventory to load vars for server1
29355 1681471136.40863: Loading VarsModule 'host_group_vars' ansible/plugins/vars/host_group_vars.py
29355 1681471136.40878:        processing dir tests/test_data/inventory/group_vars
29355 1681471136.40885: Loading data from /foo/bar
29355 1681471136.40906: Calling all_plugins_play to load vars for server1
29355 1681471136.40917: Loading VarsModule 'host_group_vars' ansible/plugins/vars/host_group_vars.py
29355 1681471136.40929: Calling groups_plugins_inventory to load vars for server1
29355 1681471136.40941: Loading VarsModule 'host_group_vars' ansible/plugins/vars/host_group_vars.py
29355 1681471136.40950:        processing dir tests/test_data/inventory/group_vars
29355 1681471136.40957: Loading data from /foo/bar.yml
29355 1681471136.40972: Calling groups_plugins_play to load vars for server1
29355 1681471136.40981: Loading VarsModule 'host_group_vars' ansible/plugins/vars/host_group_vars.py
29355 1681471136.41001: Loading VarsModule 'host_group_vars' ansible/plugins/vars/host_group_vars.py
29355 1681471136.41010:        processing dir tests/test_data/inventory/host_vars
29355 1681471136.41017: Loading data from /FOO/BAR2
29355 1681471136.41040: Loading VarsModule 'host_group_vars' ansible/plugins/vars/host_group_vars.py
29355 1681471136.41197: done with get_vars()"""

loader = DataLoader()


def test_init_variable_source():
    var_source = VariableSource(
        name="foo",
        value="bar",
        source="host vars for foobar",
        debuglog=DEBUGLOG,
    )

    assert var_source.files == ["/foo/bar", "/foo/bar.yml", "/FOO/BAR2"]
    assert var_source.source_mapped == "inventory file or script host vars"
    with pytest.raises(AnsibleFileNotFound):
        var_source.file_occurrences(loader=loader)


def test_occurrence():
    C.set_constant("CONFIG_FILE", "tests/test_data/ansible.cfg")

    var_source = VariableSource(
        name="from_all",
        value="bar",
        source="foobar",
        debuglog="""\x1b[1;30m  5108 81472141.303: Loading data from tests/test_data/inventory/group_vars/all/all\x1b[0m
        \x1b[1;30m  5108 81472141.30388: Loading data from tests/test_data/inventory/group_vars/groupA.yml\x1b[0m
        """,
    )

    assert var_source.value == "bar"
    assert var_source.source_mapped == "foobar"
    assert var_source.files == [
        "tests/test_data/inventory/group_vars/all/all",
        "tests/test_data/inventory/group_vars/groupA.yml",
    ]
    assert var_source.file_occurrences(loader=loader) == ["tests/test_data/inventory/group_vars/all/all"]


def test_empty_file():
    C.set_constant("CONFIG_FILE", "tests/test_data/ansible.cfg")

    var_source = VariableSource(
        name="from_all",
        value="bar",
        source="foobar",
        debuglog="Loading data from tests/test_data/inventory/group_vars/all/empty",
    )

    assert var_source.files == [
        "tests/test_data/inventory/group_vars/all/empty",
    ]
    assert not var_source.file_occurrences(loader=loader)



from ansible import constants as C
from ansible.inventory.manager import InventoryManager
from ansible.parsing.dataloader import DataLoader
from ansible.vars.manager import VariableManager

from ansible_variables.utils.vars import variable_sources

C.set_constant("CONFIG_FILE", "tests/test_data/ansible.cfg")

loader = DataLoader()
inventory = InventoryManager(loader=loader, sources=["tests/test_data/inventory"])
variable_manager = VariableManager(loader=loader, inventory=inventory)


def test_from_all():
    for server in ["server1", "server2", "server3"]:
        sources = variable_sources(variable_manager=variable_manager, host=inventory.get_host(server))
        assert (
            "from_all",
            "hello",
            "group vars, precedence entry 'all_plugins_inventory'",
        ) in [(source.name, source.value, source.source) for source in sources]


def test_server2():
    sources = variable_sources(variable_manager=variable_manager, host=inventory.get_host("server2"), var="test")
    assert [
        (
            "test",
            "from_groupA",
            "group vars, precedence entry 'groups_plugins_inventory'",
        )
    ] == [(source.name, source.value, source.source) for source in sources]


def test_server3():
    sources = variable_sources(variable_manager=variable_manager, host=inventory.get_host("server3"), var="test")
    assert [
        (
            "test",
            "from_groupB",
            "group vars, precedence entry 'groups_plugins_inventory'",
        )
    ] == [(source.name, source.value, source.source) for source in sources]


def test_server4():
    sources = variable_sources(variable_manager=variable_manager, host=inventory.get_host("server4"), var="test")
    assert [("test", "from_all", "group vars, precedence entry 'all_plugins_inventory'")] == [
        (source.name, source.value, source.source) for source in sources
    ]


def test_inventory_server1():
    sources = variable_sources(
        variable_manager=variable_manager, host=inventory.get_host("server1"), var="inventory_test_variable"
    )
    assert [("inventory_test_variable", "from_inventory_server1", "host vars for 'server1'")] == [
        (source.name, source.value, source.source) for source in sources
    ]


def test_inventory_server2():
    sources = variable_sources(
        variable_manager=variable_manager, host=inventory.get_host("server2"), var="inventory_test_variable"
    )
    assert [
        (
            "inventory_test_variable",
            "from_inventory_groupA",
            "group vars, precedence entry 'groups_inventory'",
        )
    ] == [(source.name, source.value, source.source) for source in sources]


def test_inventory_server3():
    sources = variable_sources(
        variable_manager=variable_manager, host=inventory.get_host("server3"), var="inventory_test_variable"
    )
    assert [
        (
            "inventory_test_variable",
            "from_inventory_all",
            "group vars, precedence entry 'all_inventory'",
        )
    ] == [(source.name, source.value, source.source) for source in sources]


def test_undefined_variable():
    sources = variable_sources(
        variable_manager=variable_manager, host=inventory.get_host("server1"), var="undefined_variable_blablabla_1234"
    )
    for variable in sources:
        assert variable.name
        assert not variable.value  # should have None value
        assert not variable.source_mapped  # should have None source



import contextlib
import io
import re
from dataclasses import dataclass
from typing import List, Optional

from ansible import constants as C
from ansible.inventory.host import Host
from ansible.parsing.dataloader import DataLoader
from ansible.utils.display import Display
from ansible.vars.manager import VariableManager, VarsWithSources

display = Display()


def escape_ansi(line):
    """The debug output contains ANSI codings, we need to remove them"""
    ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", line)


@dataclass
class VariableSource:
    """Class for keeping track of an variable source item"""

    name: str
    value: str
    source: str
    debuglog: Optional[str] = None

    @property
    def source_mapped(self) -> str:
        """Better wording of sources"""

        source_map = {
            # host variable in inventory
            "host vars for": "inventory file or script host vars",
            # group variable in inventory
            "group vars, precedence entry 'groups_inventory'": "inventory file or script group vars",
            # group variable all in inventory
            "group vars, precedence entry 'all_inventory'": "inventory file or script group vars/all",
            # host_vars
            "inventory host_vars for": "inventory host_vars/*",
            # group_vars
            "group vars, precedence entry 'groups_plugins_inventory'": "inventory group_vars/*",
            # group_vars all
            "group vars, precedence entry 'all_plugins_inventory'": "inventory group_vars/all",
        }

        for key, value in source_map.items():
            if self.source and self.source.startswith(key):
                return value

        return self.source

    @property
    def files(self) -> List[str]:
        return self.parse_files_from_debug_log()

    def parse_files_from_debug_log(self) -> List[str]:
        """The debug output from `variable_manager.get_vars()` contains all filenames
        from which the variables were loaded.

        The line looks like this:
            4890 1681462516.00300: Loading data from ansible-variables/tests/test_data/inventory/group_vars/all.yml
        """

        files = []
        if not self.debuglog:
            return files

        for line in self.debuglog.splitlines():
            found = re.search(r"Loading data from ([^\s]*)", escape_ansi(line))
            if found:
                files.extend(found.groups())

        return files

    def file_occurrences(self, loader: DataLoader):
        """Open the files and check if the variables occur in it"""
        occurrences = []

        for ffile in self.files:
            display.vvv("Checking file %s for occurrence of variable %s" % (ffile, self.name))

            # Setting unsafe=True will prevent deepcopy and will help with performance
            # and it's safe because we're not modifying the content
            content = loader.load_from_file(ffile, unsafe=True)
            if content and self.name in content:
                occurrences.append(ffile)

        return occurrences


def variable_sources(
    variable_manager: VariableManager,
    host: Host,
    var: Optional[str] = None,
) -> List[VariableSource]:
    """get vars with sources"""

    default_debug = C.DEFAULT_DEBUG
    C.DEFAULT_DEBUG = True
    # we are catching the debug messages here
    fileio = io.StringIO()
    with contextlib.redirect_stdout(fileio):
        vars_with_sources: VarsWithSources = variable_manager.get_vars(host=host)
    output = fileio.getvalue()
    C.DEFAULT_DEBUG = default_debug
    # let's print the debug message only if ANSIBLE_DEBUG was enabled before
    if C.DEFAULT_DEBUG:
        display.debug(output)

    if not var:
        return [
            VariableSource(name=var, value=value, source=vars_with_sources.get_source(var), debuglog=output)
            for var, value in vars_with_sources.items()
        ]

    return [
        VariableSource(
            name=var, value=vars_with_sources.get(var), source=vars_with_sources.get_source(var), debuglog=output
        )
    ]



import errno
import sys
import traceback
from pathlib import Path

from ansible import constants as C
from ansible import context
from ansible.cli import CLI as ACLI
from ansible.errors import AnsibleError, AnsibleOptionsError, AnsibleParserError
from ansible.module_utils._text import to_text
from ansible.utils.display import Display

display = Display()


class CLI(ACLI):
    """
    Patched `ansible.cli.CLI` class with `CLI.cli_executor` classmethod backported for ansible-core 2.11
    and ansible-core 2.12.
    Can be removed as soon as we will drop the support for ansible-core < 2.13.

    (see https://github.com/ansible/ansible/pull/76021)
    """

    @classmethod
    def cli_executor(cls, args=None):  # pylint: disable=too-many-branches,too-many-statements
        # no change for ansible-core >= 2.13
        if hasattr(super(), "cli_executor"):
            return super().cli_executor(args=args)

        # backporting code from https://github.com/ansible/ansible/blob/v2.13.9/lib/ansible/cli/__init__.py#L574
        if args is None:
            args = sys.argv

        try:
            display.debug("starting run")

            ansible_dir = Path("~/.ansible").expanduser()
            try:
                ansible_dir.mkdir(mode=0o700)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    display.warning(
                        "Failed to create the directory '%s': %s"
                        % (ansible_dir, to_text(exc, errors="surrogate_or_replace"))
                    )
            else:
                display.debug("Created the '%s' directory" % ansible_dir)

            try:
                args = [to_text(a, errors="surrogate_or_strict") for a in args]
            except UnicodeError:
                display.error(
                    """Command line args are not in utf-8, unable to continue.
                    Ansible currently only understands utf-8"""
                )
                display.display("The full traceback was:\n\n%s" % to_text(traceback.format_exc()))
                exit_code = 6
            else:
                cli = cls(args)
                exit_code = cli.run()

        except AnsibleOptionsError as exc:
            cli.parser.print_help()
            display.error(to_text(exc), wrap_text=False)
            exit_code = 5
        except AnsibleParserError as exc:
            display.error(to_text(exc), wrap_text=False)
            exit_code = 4
        # TQM takes care of these, but leaving comment to reserve the exit codes
        #    except AnsibleHostUnreachable as exc:
        #        display.error(str(exc))
        #        exit_code = 3
        #    except AnsibleHostFailed as exc:
        #        display.error(str(exc))
        #        exit_code = 2
        except AnsibleError as exc:
            display.error(to_text(exc), wrap_text=False)
            exit_code = 1
        except KeyboardInterrupt:
            display.error("User interrupted execution")
            exit_code = 99
        except Exception as exc:  # pylint: disable=broad-exception-caught
            if C.DEFAULT_DEBUG:
                # Show raw stacktraces in debug mode, It also allow pdb to
                # enter post mortem mode.
                raise
            have_cli_options = bool(context.CLIARGS)
            display.error("Unexpected Exception, this is probably a bug: %s" % to_text(exc), wrap_text=False)
            if not have_cli_options or have_cli_options and context.CLIARGS["verbosity"] > 2:
                log_only = False
                if hasattr(exc, "orig_exc"):
                    display.vvv("\nexception type: %s" % to_text(type(exc.orig_exc)))
                    why = to_text(exc.orig_exc)
                    if to_text(exc) != why:
                        display.vvv("\noriginal msg: %s" % why)
            else:
                display.display("to see the full traceback, use -vvv")
                log_only = True
            display.display("the full traceback was:\n\n%s" % to_text(traceback.format_exc()), log_only=log_only)
            exit_code = 250

        sys.exit(exit_code)



import pytest
from ansible import constants as C

from ansible_variables.cli.variables import VariablesCLI, main
from ansible_variables.utils.vars import escape_ansi

C.set_constant("CONFIG_FILE", "tests/test_data/ansible.cfg")
C.set_constant("DEFAULT_HOST_LIST", "tests/test_data/inventory")

# FIXME: we should not invoke the test with main() or VariablesCLI() as this will lead to context.CLIArgs set
# which will not be cleared between the tests and so could lead to unexpected results


def test_main(capsys):
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main(["ansible-variables"])
    assert pytest_wrapped_e.type == SystemExit
    assert "ansible-variables: error: the following arguments are required: host" in capsys.readouterr().err


def test_cli_config(capsys):
    variables_cli = VariablesCLI(["ansible-variables", "--version"])
    with pytest.raises(SystemExit):
        variables_cli.run()
    assert "tests/test_data/ansible.cfg" in capsys.readouterr().out.splitlines()[1]


def test_cli_from_all(capsys):
    for server in ["server1", "server2", "server3"]:
        variables_cli = VariablesCLI(["ansible-variables", server, "--var", "from_all"])
        variables_cli.run()
        captured = "".join(capsys.readouterr().out.splitlines())
        assert "from_all: hello - inventory group_vars/all" == escape_ansi(captured)


def test_cli_from_all_v(capsys):
    variables_cli = VariablesCLI(["ansible-variables", "server1", "--var", "from_all", "-v"])
    variables_cli.run()
    captured = "".join(capsys.readouterr().out.splitlines())
    assert "from_all: hello - inventory group_vars/all" in escape_ansi(captured)
    assert "tests/test_data/inventory/group_vars/all/all" in escape_ansi(captured)


def test_cli_from_all_vvv(capsys):
    variables_cli = VariablesCLI(["ansible-variables", "server1", "--var", "from_all", "-vvv"])
    variables_cli.run()
    captured = "".join(capsys.readouterr().out.splitlines())
    # assert "from_all: hello - inventory group_vars/all" in captured
    assert "tests/test_data/inventory/group_vars/all/all" in captured
    assert "tests/test_data/inventory/group_vars/groupA.yml" in captured
    assert "tests/test_data/inventory/group_vars/all/empty" in captured
